const utils = require('./utils');
const FORMATS = require('./formats');


// Use these to help sort formats, higher index is better.
const audioEncodingRanks = [
  'mp4a',
  'mp3',
  'vorbis',
  'aac',
  'opus',
  'flac',
];
const videoEncodingRanks = [
  'mp4v',
  'avc1',
  'Sorenson H.283',
  'MPEG-4 Visual',
  'VP8',
  'VP9',
  'H.264',
];

var getVideoBitrate = function getVideoBitrate(format) {
  return format.bitrate || 0;
};

var getVideoEncodingRank = function getVideoEncodingRank(format) {
  return videoEncodingRanks.findIndex(function (enc) {
    return format.codecs && format.codecs.includes(enc);
  });
};

var getAudioBitrate = function getAudioBitrate(format) {
  return format.audioBitrate || 0;
};

var getAudioEncodingRank = function getAudioEncodingRank(format) {
  return audioEncodingRanks.findIndex(function (enc) {
    return format.codecs && format.codecs.includes(enc);
  });
};


/**
 * Sort formats by a list of functions.
 *
 * @param {Object} a
 * @param {Object} b
 * @param {Array.<Function>} sortBy
 * @returns {number}
 */
const sortFormatsBy = function(a, b, sortBy){
  var res = 0;
  for (var fn in sortBy) {
    res = fn(b) - fn(a);
    if (res !== 0) {
      break;
    }
  }
  return res;
};


const sortFormatsByVideo = function(a, b) {
  return sortFormatsBy(a, b, [
  function(format){ return parseInt(format.qualityLabel);},
  getVideoBitrate,
  getVideoEncodingRank,
])};


const sortFormatsByAudio = function(a, b){ return sortFormatsBy(a, b, [
  getAudioBitrate,
  getAudioEncodingRank,
])};


/**
 * Sort formats from highest quality to lowest.
 *
 * @param {Object} a
 * @param {Object} b
 * @returns {number}
 */
/*
 exports.sortFormats = function(a, b){ return sortFormatsBy(a, b, [
  // Formats with both video and audio are ranked highest.
  function(format){ return +!!format.isHLS},
  function(format){ return +!!format.isDashMPD},
  function(format){ return  +(format.contentLength > 0)},
  function(format){ return  +(format.hasVideo && format.hasAudio)},
  function(format){ return  +format.hasVideo},
  function(format){ return  parseInt(format.qualityLabel) || 0},
  getVideoBitrate,
  getAudioBitrate,
  getVideoEncodingRank,
  getAudioEncodingRank,
])};
*/

var getBitrate = function(format) {return parseInt(format.bitrate) || 0};

exports.sortFormats = function(a, b) {
  var ares = a.qualityLabel ? parseInt(a.qualityLabel.slice(0, -1)) : 0;
  var bres = b.qualityLabel ? parseInt(b.qualityLabel.slice(0, -1)) : 0;
  var afeats = ~~!!ares * 2 + ~~!!a.audioBitrate;
  var bfeats = ~~!!bres * 2 + ~~!!b.audioBitrate;

  if (afeats === bfeats) {
    if (ares === bres) {
      var avbitrate = getBitrate(a);
      var bvbitrate = getBitrate(b);
      if (avbitrate === bvbitrate) {
        var aascore = audioScore(a);
        var bascore = audioScore(b);
        if (aascore === bascore) {
          var avenc = videoEncodingRanks.findIndex(function(enc) {return a.codecs && a.codecs.includes(enc)});
          var bvenc = videoEncodingRanks.findIndex(function(enc) {return b.codecs && b.codecs.includes(enc)});
          return bvenc - avenc;
        } else {
          return bascore - aascore;
        }
      } else {
        return bvbitrate - avbitrate;
      }
    } else {
      return bres - ares;
    }
  } else {
    return bfeats - afeats;
  }
};


/**
 * Choose a format depending on the given options.
 *
 * @param {Array.<Object>} formats
 * @param {Object} options
 * @returns {Object}
 * @throws {Error} when no format matches the filter/format rules
 */
exports.chooseFormat = function(formats, options){
  if (typeof options.format === 'object') {
    if (!options.format.url) {
      throw Error('Invalid format given, did you use `ytdl.getInfo()`?');
    }
    return options.format;
  }

  if (options.filter) {
    formats = exports.filterFormats(formats, options.filter);
  }

  var format;
  const quality = options.quality || 'highest';
  switch (quality) {
    case 'highest':
      format = formats[0];
      break;

    case 'lowest':
      format = formats[formats.length - 1];
      break;

    case 'highestaudio':
      formats = exports.filterFormats(formats, 'audio');
      formats.sort(sortFormatsByAudio);
      format = formats[0];
      break;

    case 'lowestaudio':
      formats = exports.filterFormats(formats, 'audio');
      formats.sort(sortFormatsByAudio);
      format = formats[formats.length - 1];
      break;

    case 'highestvideo':
      formats = exports.filterFormats(formats, 'video');
      formats.sort(sortFormatsByVideo);
      format = formats[0];
      break;

    case 'lowestvideo':
      formats = exports.filterFormats(formats, 'video');
      formats.sort(sortFormatsByVideo);
      format = formats[formats.length - 1];
      break;

    default:
      format = getFormatByQuality(quality, formats);
      break;
  }

  if (!format) {
    throw Error("No such format found: ".concat(quality));
  }
  return format;
};

/**
 * Gets a format based on quality or array of quality's
 *
 * @param {string|[string]} quality
 * @param {[Object]} formats
 * @returns {Object}
 */
const getFormatByQuality = function(quality, formats) {
  var getFormat = function getFormat(itag) {
    return formats.find(function (format) {
      return "".concat(format.itag) === "".concat(itag);
    });
  };

  if (Array.isArray(quality)) {
    return getFormat(quality.find(function (q) {
      return getFormat(q);
    }));
  } else {
    return getFormat(quality);
  }
};


/**
 * @param {Array.<Object>} formats
 * @param {Function} filter
 * @returns {Array.<Object>}
 */
exports.filterFormats = function (formats, filter) {
  var fn;

  switch (filter) {
    case 'videoandaudio':
    case 'audioandvideo':
      fn = function fn(format) {
        return format.hasVideo && format.hasAudio;
      };

      break;

    case 'video':
      fn = function fn(format) {
        return format.hasVideo;
      };

      break;

    case 'videoonly':
      fn = function fn(format) {
        return format.hasVideo && !format.hasAudio;
      };

      break;

    case 'audio':
      fn = function fn(format) {
        return format.hasAudio;
      };

      break;

    case 'audioonly':
      fn = function fn(format) {
        return !format.hasVideo && format.hasAudio;
      };

      break;

    default:
      if (typeof filter === 'function') {
        fn = filter;
      } else {
        throw TypeError("Given filter (".concat(filter, ") is not supported"));
      }

  }

  return formats.filter(function (format) {
    return !!format.url && fn(format);
  });
};



/**
 * @param {Object} format
 * @returns {Object}
 */
exports.addFormatMeta = function (format) {
  format = Object.assign({}, FORMATS[format.itag], format);
  format.hasVideo = !!format.qualityLabel;
  format.hasAudio = !!format.audioBitrate;
  format.container = format.mimeType ? format.mimeType.split(';')[0].split('/')[1] : null;
  format.codecs = format.mimeType ? utils.between(format.mimeType, 'codecs="', '"') : null;
  format.videoCodec = format.hasVideo && format.codecs ? format.codecs.split(', ')[0] : null;
  format.audioCodec = format.hasAudio && format.codecs ? format.codecs.split(', ').slice(-1)[0] : null;
  format.isLive = /\bsource[/=]yt_live_broadcast\b/.test(format.url);
  format.isHLS = /\/manifest\/hls_(variant|playlist)\//.test(format.url);
  format.isDashMPD = /\/manifest\/dash\//.test(format.url);
  return format;
};
