const urllib = require('url');
const querystring = require('querystring');
const sax = require('sax');
const miniget = require('miniget');
const utils = require('./utils');
// Forces Node JS version of setTimeout for Electron based applications
//const { setTimeout } = require('timers');

const formatUtils = require('./format-utils');
const urlUtils = require('./url-utils');
const extras = require('./info-extras');
const sig = require('./sig');
const Cache = require('./cache');
//const { format } = require('path');
//const { call } = require('../../api');


const VIDEO_URL = 'https://www.youtube.com/watch?v=';

if (!Object.assign) {
  Object.defineProperty(Object, 'assign', {
    enumerable: false,
    configurable: true,
    writable: true,
    value: function(target, firstSource) {
      'use strict';
      if (target === undefined || target === null) {
        throw new TypeError('Cannot convert first argument to object');
      }

      var to = Object(target);
      for (var i = 1; i < arguments.length; i++) {
        var nextSource = arguments[i];
        if (nextSource === undefined || nextSource === null) {
          continue;
        }

        var keysArray = Object.keys(Object(nextSource));
        for (var nextIndex = 0, len = keysArray.length; nextIndex < len; nextIndex++) {
          var nextKey = keysArray[nextIndex];
          var desc = Object.getOwnPropertyDescriptor(nextSource, nextKey);
          if (desc !== undefined && desc.enumerable) {
            to[nextKey] = nextSource[nextKey];
          }
        }
      }
      return to;
    }
  });
}

// Cached for storing basic/full info.
/*
exports.cache = new Cache();
exports.cookieCache = new Cache(1000 * 60 * 60 * 24);
exports.watchPageCache = new Cache();
*/


exports.cache = Cache;
exports.cookieCache = Cache;
exports.watchPageCache = Cache;


// Special error class used to determine if an error is unrecoverable,
// as in, ytdl-core should not try again to fetch the video metadata.
// In this case, the video is usually unavailable in some way.
const UnrecoverableError = new Error('Unrecoverable Error');


// List of URLs that show up in `notice_url` for age restricted videos.
const AGE_RESTRICTED_URLS = [
  'support.google.com/youtube/?p=age_restrictions',
  'youtube.com/t/community_guidelines',
];


/**
 * Gets info from a video without getting additional formats.
 *
 * @param {string} id
 * @param {Object} options
 * @returns {Promise<Object>}
*/

exports.getBasicInfo = function(id, options, callback){
  options = options || {};
  options.requestOptions = options.requestOptions || {};
  miniget.defaultOptions = miniget.defaultOptions || {};
  const retryOptions = Object.assign({}, miniget.defaultOptions, options.requestOptions);

  options.requestOptions = Object.assign({}, options.requestOptions, {});
  options.requestOptions.headers = Object.assign({},
    {
      // eslint-disable-next-line max-len
      'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.101 Safari/537.36',
    }, options.requestOptions.headers);
  
  const validate = function(info) {   
    var playErr = utils.playError(info.player_response, ['ERROR'], UnrecoverableError);
    var privateErr = privateVideoError(info.player_response);
    if (playErr || privateErr) {
      throw playErr || privateErr;
    }
    return info && info.player_response && (
      info.player_response.streamingData || isRental(info.player_response) || isNotYetBroadcasted(info.player_response)
    );
  };

  pipeline(id, validate, function(err, info){
      if (err) return callback(err);
      Object.assign(info, {
        formats: parseFormats(info.player_response),
        //related_videos: extras.getRelatedVideos(info),
      });

      // Add additional properties to info.
      //const media = extras.getMedia(info);
      var additional = {
        /*
        author: extras.getAuthor(info),
        media: media,
        likes: extras.getLikes(info),
        dislikes: extras.getDislikes(info),
        age_restricted: !!(media && media.notice_url && AGE_RESTRICTED_URLS.some(function(url) {return media.notice_url.includes(url)})),
        */
        // Give the standard link to the video.
        video_url: VIDEO_URL + id,

      };
    
      info.videoDetails = extras.cleanVideoDetails(Object.assign({},
        info.player_response && info.player_response.microformat &&
        info.player_response.microformat.playerMicroformatRenderer,
        info.player_response && info.player_response.videoDetails, additional));
    
      callback(null, info);
      //return info;
  });
};

const privateVideoError = function(player_response){
  var playability = player_response && player_response.playabilityStatus;
  if (playability && playability.status === 'LOGIN_REQUIRED' && playability.messages &&
    playability.messages.filter(function(m){ return /This is a private video/.test(m)}).length) {
    return new UnrecoverableError(playability.reason || (playability.messages && playability.messages[0]));
  } else {
    return null;
  }
};


const isRental = function(player_response){
  var playability = player_response.playabilityStatus;
  return playability && playability.status === 'UNPLAYABLE' &&
    playability.errorScreen && playability.errorScreen.playerLegacyDesktopYpcOfferRenderer;
};


const isNotYetBroadcasted = function(player_response){
  var playability = player_response.playabilityStatus;
  return playability && playability.status === 'LIVE_STREAM_OFFLINE';
};

const getWatchHTMLURL = function(id, options) { 
  options = options || {};
  options.lang = options.lang || 'en';
  return VIDEO_URL + id + '&hl=' + options.lang;
};
const getWatchHTMLPageBody = function(id, options, callback) {
  options = options || {};
  const url = getWatchHTMLURL(id, options);  
  var result = '';  
  miniget(url, options.requestOptions, function(err, result) {    
      if (err){
        console.log(err);
      }
      callback(result);
      //return result.text();
  });
};


const EMBED_URL = 'https://www.youtube.com/embed/';
const getEmbedPageBody = function(id, options) {
  const embedUrl = EMBED_URL + id + '?hl=' + (options.lang || 'en');
  return miniget(embedUrl, options.requestOptions).text();
};


const getHTML5player = function(body){
  var html5playerRes =
    /<script\s+src="([^"]+)"(?:\s+type="text\/javascript")?\s+name="player_ias\/base"\s*>|"jsUrl":"([^"]+)"/
      .exec(body);
  return html5playerRes ? html5playerRes[1] || html5playerRes[2] : null;
};


const getIdentityToken = function(id, options, key, throwIfNotFound) {
  return exports.cookieCache.getOrSet(key, function() {
    var page = getWatchHTMLPageBody(id, options);
    var match = page.match(/(["'])ID_TOKEN\1[:,]\s?"([^"]+)"/);
    if (!match && throwIfNotFound) {
      throw new UnrecoverableError('Cookie header used in request, but unable to find YouTube identity token');
    }
    return match && match[2];
  })};


/**
 * Goes through each endpoint in the pipeline, retrying on failure if the error is recoverable.
 * If unable to succeed with one endpoint, moves onto the next one.
 *
 * @param {Array.<Object>} args
 * @param {Function} validate
 * @param {Object} retryOptions
 * @param {Array.<Function>} endpoints
 * @returns {[Object, Object, Object]}
 */
const pipeline = function(id, validate, callback) {
    var info;
  /*
  for (var i in [0, 1, 2]) {
  //for (var func in endpoints) {
    try {
      console.log(i + ' 217');
      const newInfo = {};
      if (i == 0) {
        newInfo = getWatchHTMLPage(args[0].concat([info]));
      } 
      if (i == 1) {
        newInfo = getWatchJSONPage(args[0].concat([info]));
      } 
      if (i == 2) {
        newInfo = getVideoInfoPage(args[0].concat([info]));
      } 
      //const newInfo = retryFunc(func, args.concat([info]), retryOptions);
      if (newInfo.player_response) {
          newInfo.player_response.videoDetails = assign(
          info && info.player_response && info.player_response.videoDetails,
          newInfo.player_response.videoDetails);
        newInfo.player_response = assign(info && info.player_response, newInfo.player_response);
      }
      info = assign(info, newInfo);
      if (validate(info, false)) {
        break;
      }
    } catch (err) {
      //if (err instanceof UnrecoverableError || func === endpoints[endpoints.length - 1]) {      
        console.log('Error newInfo 253');
        throw err;
      //}
      // Unable to find video metadata... so try next endpoint.
    }
  } */    
    getWatchHTMLPage(id, function(err, newInfo) {
        if (err) { 
            callback(err);
        }
        if (newInfo.player_response) {
          newInfo.player_response.videoDetails = assign(
          info && info.player_response && info.player_response.videoDetails,
          newInfo.player_response.videoDetails);
        newInfo.player_response = assign(info && info.player_response, newInfo.player_response);
        }
        info = assign(info, newInfo);
        callback(null, info);
      });
  //return info;
};



/**
 * Like Object.assign(), but ignores `null` and `undefined` from `source`.
 *
 * @param {Object} target
 * @param {Object} source
 * @returns {Object}
 */
const assign = function(target, source) {
  if (!target || !source) { return target || source; }
  //for (var [key, value] in Object.entries(source)) {
  for (var key in source) {  
    var value = source[key];
    if (value !== null && value !== undefined) {
      target[key] = value;
    }
  }
  return target;
};


/**
 * Given a function, calls it with `args` until it's successful,
 * or until it encounters an unrecoverable error.
 * Currently, any error from miniget is considered unrecoverable. Errors such as
 * too many redirects, invalid URL, status code 404, status code 502.
 *
 * @param {Function} func
 * @param {Array.<Object>} args
 * @param {Object} options
 * @param {number} options.maxRetries
 * @param {Object} options.backoff
 * @param {number} options.backoff.inc
 */
const retryFunc = function(func, args, options) {
  var currentTry = 0, result;
  while (currentTry <= options.maxRetries) {
    try {
      result = func(args);
      break;
    } catch (err) {
      if (err instanceof UnrecoverableError ||
        (err instanceof miniget.MinigetError && err.statusCode < 500) || currentTry >= options.maxRetries) {
        throw err;
      }
      var wait = Math.min(++currentTry * options.backoff.inc, options.backoff.max);
      function ret(){
        console.log('retry... '+ wait);
      }
      setTimeout(ret, wait);
    }
  }
  return result;
};


const jsonClosingChars = /^[)\]}'\s]+/;
const parseJSON = function(source, varName, json) {
  if (!json || typeof json === 'object') {
    return json;
  } else {
    try {
      json = json.replace(jsonClosingChars, '');
      return JSON.parse(json);
    } catch (err) {
      throw Error('Error parsing ' + varName +' in ' + source + ': ' + err.message);
    }
  }
};


const findJSON = function(source, varName, body, left, right, prependJSON){
  prependJSON = typeof prependJSON !== 'undefined' ?  prependJSON : '';
  var jsonStr = utils.between(body, left, right);
  if (!jsonStr) {
    throw Error('Could not find ' + varName + ' in ' + source);
  }
  return parseJSON(source, varName, utils.cutAfterJSON(prependJSON + jsonStr));
};


const findPlayerResponse = function(source, info) {
  const player_response = info && (
    (info.args && info.args.player_response) ||
    info.player_response || info.playerResponse || info.embedded_player_response);
  return parseJSON(source, 'player_response', player_response);
};


const getWatchJSONURL = function(id, options) { return getWatchHTMLURL(id, options) + '&pbj=1'};
const getWatchJSONPage = function(id, options) {
  const reqOptions = Object.assign({ headers: {} }, options.requestOptions);
  var cookie = reqOptions.headers.Cookie || reqOptions.headers.cookie;
  reqOptions.headers = Object.assign({
    'x-youtube-client-name': '1',
    'x-youtube-client-version': '2.20201203.06.00',
    'x-youtube-identity-token': exports.cookieCache.get(cookie || 'browser') || '',
  }, reqOptions.headers);

  const setIdentityToken = function(key, throwIfNotFound) {
    if (reqOptions.headers['x-youtube-identity-token']) { return; }
    reqOptions.headers['x-youtube-identity-token'] = getIdentityToken(id, options, key, throwIfNotFound);
  };

  if (cookie) {
    setIdentityToken(cookie, true);
  }

  const jsonUrl = getWatchJSONURL(id, options);
  var body = miniget(jsonUrl, reqOptions).text();
  var parsedBody = parseJSON('watch.json', 'body', body);
  if (parsedBody.reload === 'now') {
    setIdentityToken('browser', false);
  }
  if (parsedBody.reload === 'now' || !Array.isArray(parsedBody)) {
    throw Error('Unable to retrieve video metadata in watch.json');
  }
  var info = parsedBody.reduce(function(part, curr) { return Object.assign(curr, part), {}});
  info.player_response = findPlayerResponse('watch.json', info);
  info.html5player = info.player && info.player.assets && info.player.assets.js;

  return info;
};


const getWatchHTMLPage = function(id, callback) {
  //const getWatchHTMLPage = function(id, options, callback) {
  //options = options || {};
  options = {};
  if (typeof(callback) !== 'function'){
      console.log('callback Not function');
  }
  getWatchHTMLPageBody(id, options, function(body){
    var info = { page: 'watch' };
    try {
      info.player_response = findJSON('watch.html', 'player_response',
        body, /\bytInitialPlayerResponse\s*=\s*\{/i, '</script>', '{');
    } catch (err) {
      var args = findJSON('watch.html', 'player_response', body, /\bytplayer\.config\s*=\s*{/, '</script>', '{');
      info.player_response = findPlayerResponse('watch.html', args);
    }
    info.response = findJSON('watch.html', 'response', body, /\bytInitialData("\])?\s*=\s*\{/i, '</script>', '{');
    info.html5player = getHTML5player(body);
    callback(null, info);
  });

};


const INFO_HOST = 'www.youtube.com';
const INFO_PATH = '/get_video_info';
const VIDEO_EURL = 'https://youtube.googleapis.com/v/';
const getVideoInfoPage = function(id, options) {
  const url = urllib.format({
    protocol: 'https',
    host: INFO_HOST,
    pathname: INFO_PATH,
    query: {
      video_id: id,
      eurl: VIDEO_EURL + id,
      ps: 'default',
      gl: 'US',
      hl: options.lang || 'en',
    },
  });
  var body = miniget(url, options.requestOptions).text();
  var info = querystring.parse(body);
  info.player_response = findPlayerResponse('get_video_info', info);
  return info;
};


/**
 * @param {Object} player_response
 * @returns {Array.<Object>}
 */
const parseFormats = function(player_response) {
  var formats = [];
  if (player_response && player_response.streamingData) {    
    formats = formats
      .concat(player_response.streamingData.formats || [])
      .concat(player_response.streamingData.adaptiveFormats || []);  
//    formats = formats.concat(player_response.streamingData.formats).concat(player_response.streamingData.adaptiveFormats);  
    }
  return formats;
};


/**
 * Gets info from a video additional formats and deciphered URLs.
 *
 * @param {string} id
 * @param {Object} options
 * @returns {Promise<Object>}
 */
exports.getInfo = function(id, options, callback){
  if (typeof options === 'function') {
    callback = options;
    options = {};
  } else if (!options) {
    options = {};
  }
  return exports.getBasicInfo(id, options, function(err, info) {
  //var info = exports.getBasicInfo(id, options);
  if (err) return callback(err);
  const hasManifest =
    info.player_response && info.player_response.streamingData && (
      info.player_response.streamingData.dashManifestUrl ||
      info.player_response.streamingData.hlsManifestUrl
    );
    
  //var funcs = [];

  if (info.formats.length) {
    info.html5player = info.html5player ||
      getHTML5player(getWatchHTMLPageBody(id, options)) || getHTML5player(getEmbedPageBody(id, options));
    if (!info.html5player) {
      throw Error('Unable to find html5player file');
    }
    const html5player = urllib.resolve(VIDEO_URL, info.html5player);
    //funcs.push(sig.decipherFormats(info.formats, html5player, options));    
    sig.decipherFormats(info.formats, html5player, options.debug, function(decipheredForm){
      info.formats = info.formats.map(formatUtils.addFormatMeta);
      info.formats.sort(formatUtils.sortFormats);      
      info.full = true;
      callback(null, info);  
    });
  }else{
    console.error('This video is unavailable');
  }   
  /*
  if (hasManifest && info.player_response.streamingData.dashManifestUrl) {
    var url = info.player_response.streamingData.dashManifestUrl;
    funcs.push(getDashManifest(url, options));
  }
  if (hasManifest && info.player_response.streamingData.hlsManifestUrl) {
    var url = info.player_response.streamingData.hlsManifestUrl;
    funcs.push(getM3U8(url, options));
  }
  var results = funcs;
  info.formats = Object.values(Object.assign({}, results));
  */
  //return info;
  });
}


/**
 * Gets additional DASH formats.
 *
 * @param {string} url
 * @param {Object} options
 * @returns {Promise<Array.<Object>>}
 */
const getDashManifest = function(url, options) {
  var formats = {};
  const parser = sax.parser(false);
  parser.onerror = reject;
  var adaptationSet;
  parser.onopentag = function(node) {
    if (node.name === 'ADAPTATIONSET') {
      adaptationSet = node.attributes;
    } else if (node.name === 'REPRESENTATION') {
      const itag = parseInt(node.attributes.ID);
      if (!isNaN(itag)) {
        formats[url] = Object.assign({
          itag: itag, 
          url: url,
          bitrate: parseInt(node.attributes.BANDWIDTH),
          mimeType: adaptationSet.MIMETYPE + '; codecs="' + node.attributes.CODECS + '"',
        }, node.attributes.HEIGHT ? {
          width: parseInt(node.attributes.WIDTH),
          height: parseInt(node.attributes.HEIGHT),
          fps: parseInt(node.attributes.FRAMERATE),
        } : {
          audioSampleRate: node.attributes.AUDIOSAMPLINGRATE,
        });
      }
    }
  };
  parser.onend = function() { resolve(formats); };
  const req = miniget(urllib.resolve(VIDEO_URL, url), options.requestOptions);
  req.setEncoding('utf8');
  req.on('error', reject);
  req.on('data', function(chunk) { return parser.write(chunk); });
  req.on('end', parser.close.bind(parser));
};


/**
 * Gets additional formats.
 *
 * @param {string} url
 * @param {Object} options
 * @returns {Promise<Array.<Object>>}
 */
const getM3U8 = function(url, options) {
  url = urllib.resolve(VIDEO_URL, url);
  var body = miniget(url, options.requestOptions).text();
  var formats = {};
  body
    .split('\n')
    .filter(function(line) { return /^https?:\/\//.test(line)})
    .forEach(function(line) {
      const itag = parseInt(line.match(/\/itag\/(\d+)\//)[1]);
      formats[line] = { itag: itag, url: line };
    });
  return formats;
};


// Cache get info functions.
// In case a user wants to get a video's info before downloading.
for (var funcName in ['getBasicInfo', 'getInfo']) {
  /**
   * @param {string} link
   * @param {Object} options
   * @returns {Promise<Object>}
   */
  const func = exports[funcName];
  exports[funcName] = function(link, options) {
    options = typeof options !== 'undefined' ?  options : {};
    //utils.checkForUpdates();
    var id = urlUtils.getVideoID(link);
    const key = [funcName, id, options.lang].join('-');
    return exports.cache.getOrSet(key, function(){ return func(id, options)});
  };
}


// Export a few helpers.
exports.validateID = urlUtils.validateID;
exports.validateURL = urlUtils.validateURL;
exports.getURLVideoID = urlUtils.getURLVideoID;
exports.getVideoID = urlUtils.getVideoID;
