const querystring = require('querystring');
// const Cache = require('./cache');
const urllib = require('url');
const miniget = require('miniget');
const utils = require('./utils');


// A shared cache to keep track of html5player js functions.
// exports.cache = Cache();

/**
 * Extract signature deciphering and n parameter transform functions from html5player file.
 *
 * @param {string} html5playerfile
 * @param {Object} options
 * @returns {Promise<Array.<string>>}
 */
exports.getFunctions = function(html5playerfile, options,  callback) {
  options = options || {};
  miniget(html5playerfile, options, function(err, body){
    const functions = exports.extractFunctions(body);
    if (!functions || !functions.length) {
      throw Error('Could not extract functions');
    }
    callback(functions);
  });
}

/**
 * Extracts the actions that should be taken to decipher a signature
 * and tranform the n parameter
 *
 * @param {string} body
 * @returns {Array.<string>}
 */
exports.extractFunctions = function(body){
  const functions = [];
  const extractManipulations = function(caller){    
    var functionName = utils.between(caller, 'a=a.split("");', '.');
    if (!functionName) return '';
    const functionStart = 'var ' + functionName + '={';
    const ndx = body.indexOf(functionStart);
    if (ndx < 0) return '';
    const subBody = body.slice(ndx + functionStart.length - 1);
    return 'var ' + functionName + '=' + utils.cutAfterJSON(subBody);
  };
  const extractDecipher = function(){
    const functionName = utils.between(body, 'a.set("alr","yes");c&&(c=', '(decodeURIC');
    if (functionName && functionName.length) {
      const functionStart = functionName + '=function(a)';
      const ndx = body.indexOf(functionStart);
      if (ndx >= 0) {
        const subBody = body.slice(ndx + functionStart.length);
        var functionBody = 'var ' + functionStart + utils.cutAfterJSON(subBody);
        functionBody = extractManipulations(functionBody) + ';' + functionBody + ';' + functionName + '(sig);';
        functions.push(functionBody);
      }
    }
  };
  const extractNCode = function(){
    const functionName = utils.between(body, '&&(b=a.get("n"))&&(b=', '(b)');
    if (functionName && functionName.length) {
      const functionStart = functionName + '=function(a)';
      const ndx = body.indexOf(functionStart);
      if (ndx >= 0) {
        const subBody = body.slice(ndx + functionStart.length);
        const functionBody = 'var ' + functionStart + utils.cutAfterJSON(subBody) + ';' + functionName + '(ncode);';
        functions.push(functionBody);
      }
    }
  };
  extractDecipher();
  extractNCode();
  return functions;
};

/**
 * Apply decipher and n-transform to individual format
 *
 * @param {Object} format
 * @param {vm.Script} decipherScript
 * @param {vm.Script} nTransformScript
 */
exports.setDownloadURL = function(format, decipherScript, nTransformScript){
  const decipher = function(url){
    const args = querystring.parse(url);
    if (!args.s || !decipherScript) return args.url;
    const components = urllib.parse(decodeURIComponent(args.url), true);
    delete components.search;
    const sig = decodeURIComponent(args.s);
    components.query[args.sp ? args.sp : 'signature'] = 
      eval(decipherScript);
    return urllib.format(components);
  };
  const ncode = function(url){
    const components = urllib.parse(decodeURIComponent(url), true);
    delete components.search;
    const n = components.query['n'];
    if (!n || !nTransformScript) return url;
    const ncode = n;
    components.query['n'] =  eval(nTransformScript);
    return urllib.format(components);
  };
  const cipher = !format.url;
  const url = format.url || format.signatureCipher || format.cipher;
  format.url = cipher ? ncode(decipher(url)) : ncode(url);
  delete format.signatureCipher;
  delete format.cipher;
};

/**
 * Applies decipher and n parameter transforms to all format URL's.
 *
 * @param {Array.<Object>} formats
 * @param {string} html5player
 * @param {Object} options
 */
exports.decipherFormats = function(formats, html5player, options, callback){
  var decipheredFormats = {};
  exports.getFunctions(html5player, options, function(functions){  
    const decipherScript = functions.length ? functions[0] : null;
    const nTransformScript = functions.length > 1 ? functions[1] : null;
    formats.forEach(function(format){
      exports.setDownloadURL(format, decipherScript, nTransformScript);
      decipheredFormats[format.url] = format;
    });
   callback(decipheredFormats);
  });
};
